import os
import pathlib
from abc import ABC, abstractmethod

from PIL import Image

import av

class MediaInfoAbstract(ABC):
	filename = ""
	container = None
	duration = 0.0

	def __init__(self, filename: str) -> None:
		self.filename = filename
		self.container = av.open(file=self.filename, mode='r')

	def __del__(self) -> None:
		if self.container:
			self.container.close()

	def get_duration(self) -> float:
		return self.duration

	@staticmethod
	def get_filesize(filename: str) -> float:
		return os.stat(filename).st_size

	@abstractmethod
	def get_finfo(cls, except_info: tuple=()) -> dict:
		raise NotImplementedError

	@staticmethod
	def shrink_image_to_fit(image: Image, size: tuple = (320, 320)) -> Image:
		image.thumbnail(size, Image.Resampling.LANCZOS)
		#image.save(
		#	"/tmp/test.th.jpg",
		#	quality=80,
		#)
		return image

	def generate_filepath(self, base_filepath: str, postfix: str = "silenced") -> str:
		path_info = pathlib.Path(base_filepath)
		ext = path_info.suffix
		old_filename = path_info.stem
		new_filename = "%s_%s%s" % (old_filename, postfix, ext)
		new_filepath = "%s/%s" % (os.path.dirname(base_filepath), new_filename)

		return new_filepath



import os
import io
import pathlib
import time
import math
import socket
import ssl
#from abc import abstractmethod
from typing import Callable, Union, Optional
import logging
import json
import http.client
import pytubefix.exceptions
import requests
from PIL import Image
import numpy as np
import urllib
import urllib3
from urllib.parse import urlparse, parse_qs

import pytubefix
from pytubefix import YouTube
from pytubefix.innertube import _default_clients
from pytubefix.streams import Stream
#from pytubefix.exceptions import VideoUnavailable, VideoPrivate, MaxRetriesExceeded
import yt_dlp

from warp_beacon.jobs.download_job import DownloadJob
from warp_beacon.scraper.abstract import ScraperAbstract
from warp_beacon.yt_auth import YtAuth
from warp_beacon.scraper.exceptions import TimeOut, Unavailable, BadProxy, extract_exception_message

class YoutubeAbstract(ScraperAbstract):
	DOWNLOAD_DIR = "/tmp"
	YT_SESSION_FILE = '/var/warp_beacon/yt_session_%d.json'

	def validate_session(self) -> int:
		try:
			logging.info("Validating YT session(s) ...")
			session_dir = os.path.dirname(self.YT_SESSION_FILE)
			for f in os.listdir(session_dir):
				if f.startswith("yt_session") and f.endswith(".json"):
					yt_sess_file = f"{session_dir}/{f}"
					if os.path.exists(yt_sess_file):
						account_index = int(f.split('_')[-1].rstrip('.json'))
						logging.info("Validating YT session #%d ...", account_index)
						yt_sess_data, exp = {}, 0
						with open(yt_sess_file, 'r', encoding="utf-8") as f:
							yt_sess_data = json.loads(f.read())
							exp = int(yt_sess_data.get("expires", 0))
						if exp <= time.time() + 60:
							yt_auth = YtAuth(account_index=account_index)
							requests_data = yt_auth.refresh_token(refresh_token=yt_sess_data.get("refresh_token", ""))
							if requests_data:
								yt_sess_data.update(requests_data)
								if yt_auth.safe_write_session(yt_sess_data):
									logging.info("YT session #%d validated", account_index)
		except Exception as e:
			logging.error("Failed to refresh Youtube session!")
			logging.exception(e)

		return 0

	def rename_local_file(self, filename: str) -> str:
		if not os.path.exists(filename):
			raise NameError("No file provided")
		path_info = pathlib.Path(filename)
		ext = path_info.suffix
		#old_filename = path_info.stem
		time_name = str(time.time()).replace('.', '_')
		new_filename = "%s%s" % (time_name, ext)
		new_filepath = "%s/%s" % (os.path.dirname(filename), new_filename)

		os.rename(filename, new_filepath)

		return new_filepath
	
	def get_video_id(self, url: str) -> Optional[str]:
		parsed_url = urlparse(url)
		query = parse_qs(parsed_url.query)
		return query.get('v', [None])[0]

	def remove_tmp_files(self) -> None:
		for i in os.listdir(self.DOWNLOAD_DIR):
			if "yt_download_" in i:
				os.unlink("%s/%s" % (self.DOWNLOAD_DIR, i))

	def calculate_size_with_padding(self,
		image: Image,
		aspect_ratio_width: int,
		aspect_ratio_height: int,
		target_size: tuple=(320, 320),
		background_color: tuple=(0, 0, 0)
	) -> Image:
		aspect_ratio = aspect_ratio_width / aspect_ratio_height
		target_width, target_height = target_size

		height = target_height
		width = int(height * aspect_ratio)

		if width > target_width:
			width = target_width
			height = int(width / aspect_ratio)

		new_image = Image.new("RGB", target_size, background_color)
		image.thumbnail(target_size, Image.Resampling.LANCZOS)

		#if aspect_ratio_height > 4:
		#	image = image.resize((image.size[0], image.size[1]+new_image.size[1]))
		#	target_height += new_image.size[1] - 5

		paste_position = ((target_width - width) // 2, (target_height - height) // 2)
		new_image.paste(image, paste_position)

		return new_image

	def resize_aspect_ratio(self, image: Image, base_width: int = 320) -> Image:
		wpercent = (base_width / float(image.size[0]))
		hsize = int((float(image.size[1]) * float(wpercent)))
		img = image.resize((base_width, hsize), Image.Resampling.LANCZOS)

		return img

	def aspect_ratio(self, size: tuple) -> tuple:
		gcd = math.gcd(size[0], size[1])
		return size[0] // gcd, size[1] // gcd
	
	def crop_black_edges_pil(self, img: Image, threshold: int = 35) -> Image:
		y_nonzero, x_nonzero, _ = np.nonzero(np.array(img) > threshold)
		return img.crop((np.min(x_nonzero), np.min(y_nonzero), np.max(x_nonzero), np.max(y_nonzero)))

	def crop_center(self, image: Image, width: int, height: int) -> Image:
		img_width, img_height = image.size
		left = (img_width - width) // 2
		top = (img_height - height) // 2
		right = left + width
		bottom = top + height

		return image.crop((left, top, right, bottom))

	def download_thumbnail(self, video_id: str, timeout: int, crop_center: dict = None) -> Optional[io.BytesIO]:
		for i in ("https://img.youtube.com/vi/{VIDEO_ID}/maxresdefault.jpg",
				"https://img.youtube.com/vi/{VIDEO_ID}/hqdefault.jpg",
				"https://img.youtube.com/vi/{VIDEO_ID}/sddefault.jpg"):
			try:
				url = i.format(VIDEO_ID=video_id)
				logging.info("Youtube thumbnail url '%s'", url)
				proxies = None
				if self.proxy:
					dsn = self.proxy.get("dsn", None)
					if dsn:
						proxies = self.build_proxies(dsn)
				with requests.get(url, timeout=(timeout, timeout), proxies=proxies) as response:
					if response.status_code == 200:
						image = Image.open(io.BytesIO(response.content))
						ratio = self.aspect_ratio(image.size)
						image = self.crop_black_edges_pil(image)
						if crop_center:
							image = self.crop_center(image, width=crop_center["width"], height=crop_center["height"])
						logging.info("thumb ratio: '%s'", ratio)
						new_image = self.resize_aspect_ratio(image)
						logging.info("thumb size: '%s'", new_image.size)
						io_buf = io.BytesIO()
						new_image.save(io_buf, format='JPEG', subsampling=0, quality=95, progressive=True, optimize=False)
						logging.info("thumb size: %d kb", io_buf.getbuffer().nbytes / 1024)
						io_buf.seek(0)
						return io_buf
			except Exception as e:
				logging.error("Failed to download download thumbnail!")
				logging.exception(e)

		return None

	def download_hndlr(self, func: Callable, *args: tuple[Union[str, int, dict, tuple, bool]], **kwargs: dict[Union[str, int, dict, tuple, bool]]) -> Optional[Union[list, dict, str, io.BytesIO]]:
		ret_val = None
		max_retries = int(os.environ.get("YT_MAX_RETRIES", default=self.YT_MAX_RETRIES_DEFAULT))
		pause_secs = int(os.environ.get("YT_PAUSE_BEFORE_RETRY", default=self.YT_PAUSE_BEFORE_RETRY_DEFAULT))
		timeout = int(os.environ.get("YT_TIMEOUT", default=self.YT_TIMEOUT_DEFAULT))
		timeout_increment = int(os.environ.get("YT_TIMEOUT_INCREMENT", default=self.YT_TIMEOUT_INCREMENT_DEFAULT))
		retries = 0
		while max_retries >= retries:
			try:
				kwargs["timeout"] = timeout
				ret_val = func(*args, **kwargs)
				break
			except pytubefix.exceptions.MaxRetriesExceeded:
				# do noting, not interested
				pass
			#except http.client.IncompleteRead as e:
			except (KeyError, pytubefix.exceptions.RegexMatchError):
				raise Unavailable("Library failed")
			except urllib3.exceptions.ProxyError as e:
				logging.warning("Proxy error!")
				raise BadProxy(extract_exception_message(e.original_error))
			except (socket.timeout,
					ssl.SSLError,
					http.client.IncompleteRead,
					http.client.HTTPException,
					requests.RequestException,
					urllib.error.URLError,
					urllib.error.HTTPError) as e:
				if hasattr(e, "code") and (int(e.code) == 403 or int(e.code) == 400):
					raise Unavailable(extract_exception_message(e))
				if hasattr(e, "reason") and "Remote end closed connection without response" in str(e.reason):
					raise Unavailable(extract_exception_message(e))
				logging.warning("Youtube read timeout! Retrying in '%d' seconds ...", pause_secs)
				logging.info("Your `YT_MAX_RETRIES` values is '%d'", max_retries)
				logging.exception(extract_exception_message(e))
				if max_retries <= retries:
					#self.remove_tmp_files()
					raise TimeOut(extract_exception_message(e))
				retries += 1
				timeout += timeout_increment
				time.sleep(pause_secs)
			except (pytubefix.exceptions.VideoUnavailable, pytubefix.exceptions.VideoPrivate) as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.DownloadError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.GeoRestrictedError:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.PostProcessingError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.ExtractorError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.MaxDownloadsReached as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.UnavailableVideoError as e:
				raise Unavailable(extract_exception_message(e))
			except yt_dlp.utils.ThrottledDownload as e:
				raise Unavailable(extract_exception_message(e))

		return ret_val

	def yt_on_progress(self, stream: Stream, chunk: bytes, bytes_remaining: int) -> None:
		pass
		#logging.info("bytes: %d, bytes remaining: %d", chunk, bytes_remaining)

	def build_proxies(self, proxy_dsn: str) -> dict:
		if not proxy_dsn:
			logging.warning("Empty DSN!")
			return {}
		proxy = {}
		if proxy_dsn:
			if "https://" in proxy_dsn:
				proxy["https"] = proxy_dsn
			elif "http://" in proxy_dsn:
				proxy["http"] = proxy_dsn
			else:
				logging.warning("Proxy DSN malformed!")

		return proxy

	def build_yt(self, url: str, session: bool = True) -> YouTube:
		_default_clients["ANDROID"]["innertube_context"]["context"]["client"]["clientVersion"] = "19.08.35"
		_default_clients["ANDROID_MUSIC"] = _default_clients["ANDROID"]
		yt_opts = {"url": url, "on_progress_callback": self.yt_on_progress}
		if session:
			yt_opts["client"] = "TV_EMBED"
			yt_opts["use_oauth"] = True
			yt_opts["allow_oauth_cache"] = True
			yt_opts["token_file"] = self.YT_SESSION_FILE % self.account_index
			if not os.path.exists(yt_opts["token_file"]):
				logging.warning("YT session '%s' file is not found", yt_opts["token_file"])
				self.request_yt_auth()
				self.auth_event.wait()
				yt_auth = YtAuth(account_index=self.account_index)
				device_code = yt_auth.load_device_code()
				if device_code:
					auth_data = yt_auth.confirm_token(device_code=device_code)
					if auth_data:
						yt_auth.safe_write_session(auth_data)
				else:
					logging.error("Failed to fetch YT auth token!")
		if self.proxy:
			proxy_dsn = self.proxy.get("dsn", "")
			logging.info("Using proxy DSN '%s'", proxy_dsn)
			if proxy_dsn:
				yt_opts["proxies"] = self.build_proxies(proxy_dsn)
		return YouTube(**yt_opts)
	
	def build_yt_dlp(self, timeout: int = 60) -> yt_dlp.YoutubeDL:
		auth_data = {}
		with open(self.YT_SESSION_FILE % self.account_index, 'r', encoding="utf-8") as f:
			auth_data = json.loads(f.read())
		time_name = str(time.time()).replace('.', '_')
		ydl_opts = {
			'socket_timeout': timeout,
			'outtmpl': f'{self.DOWNLOAD_DIR}/yt_download_{time_name}.%(ext)s',
			'format': 'bestvideo+bestaudio/best',
			'merge_output_format': 'mp4',
			'noplaylist': True,
			'tv_auth': auth_data
		}

		if self.proxy:
			proxy_dsn = self.proxy.get("dsn", "")
			logging.info("Using proxy DSN '%s'", proxy_dsn)
			if proxy_dsn:
				ydl_opts["proxy"] = proxy_dsn

		return yt_dlp.YoutubeDL(ydl_opts)
	
	def _download(self, _: str, timeout: int = 60) -> list:
		raise NotImplementedError("You should to implement _download method")
	
	def _download_yt_dlp(self, _: str, timeout: int = 60) -> list:
		raise NotImplementedError("You should to implement _download_yt_dlp method")

	def download(self, job: DownloadJob) -> list:
		ret = []
		thumbnail = None
		try:
			video_id = self.get_video_id(job.url)
			# shorts custom thumb
			##vinfo = VideoInfo(local_file)
			#thumbnail = self.download_hndlr(self.download_thumbnail, video_id=yt.video_id, crop_center=vinfo.get_demensions())
			if video_id:
				thumbnail = self.download_hndlr(self.download_thumbnail, video_id)
		except Exception as e:
			logging.error("Failed to download thumb!")
			logging.exception(e)

		try:
			ret = self.download_hndlr(self._download, job.url, session=True, thumbnail=thumbnail)
			return ret
		except (Unavailable, TimeOut, KeyError) as e:
			logging.warning("Download failed, trying to download with yt_dlp")
			logging.exception(e)
		
		try:
			ret = self.download_hndlr(self._download_yt_dlp, job.url, thumbnail=thumbnail)
		except NotImplementedError:
			logging.info("yt_dlp is not supported for this submodule yet")
			raise Unavailable("Сontent unvailable")

		return ret



import os
import time
import random
import json
import re
from typing import Optional, List

from itertools import cycle
import multiprocessing
import multiprocessing.managers

from warp_beacon.jobs import Origin

import logging

class AccountSelector(object):
	accounts = []
	proxies = []
	current = None
	current_module_name = None
	accounts_meta_data = None
	session_dir = "/var/warp_beacon"
	manager = None
	account_index = {}
	current_proxy = None
	ig_request_count = None

	def __init__(self, manager: multiprocessing.managers.SyncManager, acc_file_path: str, proxy_file_path: str=None) -> None:
		self.manager = manager
		self.accounts_meta_data = self.manager.dict()
		if os.path.exists(acc_file_path):
			with open(acc_file_path, 'r', encoding="utf-8") as f:
				self.accounts = json.loads(f.read())
			if self.accounts:
				self.ig_request_count = self.manager.Value('i', 0)
				self.__init_meta_data()
				#self.load_yt_sessions()
				for acc_type, _ in self.accounts.items():
					self.account_index[acc_type] = self.manager.Value('i', 0)
			if proxy_file_path:
				with open(proxy_file_path, 'r', encoding="utf-8") as f:
					self.proxies = json.loads(f.read())

			self.load_ig_request_count()
		else:
			raise ValueError("Accounts file not found")

	def save_ig_request_count(self) -> None:
		try:
			state = {"ig_count": self.ig_request_count.value}
			with open(f"{self.session_dir}/ig_request_counter.json", "w+", encoding="utf-8") as f:
				f.write(json.dumps(state))
		except Exception as e:
			logging.error("Failed to save accounts states!")
			logging.exception(e)

	def load_ig_request_count(self) -> None:
		try:
			state = {}
			filepath = f"{self.session_dir}/ig_request_counter.json"
			if os.path.exists(filepath):
				with open(f"{self.session_dir}/ig_request_counter.json", "r", encoding="utf-8") as f:
					state = json.loads(f.read())
				if state:
					self.ig_request_count.value = int(state.get("ig_count", 0))
		except Exception as e:
			logging.error("Failed to save accounts states!")
			logging.exception(e)

	def get_current_proxy(self) -> Optional[dict]:
		return self.current_proxy
	
	def get_last_proxy(self) -> Optional[dict]:
		return self.accounts_meta_data.get("last_proxy", None)

	def get_proxy_list(self) -> List[dict]:
		matched_proxy = []
		try:
			acc_id, acc_data = self.get_current()
			current_acc_pid = acc_data.get("proxy_id", "").strip()
			for proxy in self.proxies:
				pid = proxy.get("id", "").strip()
				if pid and current_acc_pid and pid == current_acc_pid:
					if "override_force_ipv6" in proxy:
						self.accounts[self.current_module_name][acc_id]["force_ipv6"] = proxy.get("override_force_ipv6", False)
					logging.info("Account proxy matched '%s'", proxy)
					matched_proxy.append(proxy)
		except Exception as e:
			logging.warning("Failed to form proxy list!")
			logging.exception(e)

		return matched_proxy

	def get_random_account_proxy(self) -> Optional[dict]:
		if self.proxies:
			try:
				matched_proxy = self.get_proxy_list()
				if matched_proxy:
					if len(matched_proxy) > 1:
						random.seed(random.seed(time.time_ns() ^ int.from_bytes(os.urandom(len(matched_proxy)), "big")))
						# ensure new proxy in case if previous account required captcha
						last_proxy = self.get_last_proxy()
						if last_proxy and last_proxy in matched_proxy:
							matched_proxy.remove(last_proxy)
					prox_choice = random.choice(matched_proxy)
					# saving chosen proxy for history
					self.accounts_meta_data["last_proxy"] = prox_choice
					self.current_proxy = prox_choice
					logging.info("Chosen proxy: '%s'", prox_choice)
					return prox_choice
			except Exception as e:
				logging.warning("Error on selecting account proxy!")
				logging.exception(e)
		return None
	
	def next_proxy(self) -> Optional[dict]:
		if not self.proxies:
			return None
		proxy = None
		try:
			matched_proxies = self.get_proxy_list()
			if matched_proxies:
				lit = cycle(matched_proxies)
				proxy = next(lit)
				last_proxy = self.get_last_proxy()
				if last_proxy and proxy.get("dsn", "") == last_proxy.get("dsn", ""):
					proxy = next(lit)
				self.current_proxy = proxy
				self.accounts_meta_data["last_proxy"] = proxy
		except Exception as e:
			logging.warning("Error on selection next proxy!")
			logging.exception(e)

		return proxy

	def load_yt_sessions(self) -> None:
		if "youtube" not in self.accounts:
			self.accounts["youtube"] = []
		for f in os.listdir(self.session_dir):
			if "yt_session" in f and ".json" in f:
				match = re.search(r'\d+', f)
				index = 0
				if match:
					index = int(match.group(0))
				self.accounts["youtube"].insert(index, {"session_file": "%s/%s" % (self.session_dir, f), "index": index})
		if not self.accounts["youtube"]:
			self.accounts["youtube"].append({"session_file": "%s/yt_session_0.json" % self.session_dir, "index": 0})

	def __init_meta_data(self) -> None:
		for module_name, lst in self.accounts.items():
			if module_name not in self.accounts_meta_data:
				self.accounts_meta_data[module_name] = []
			for index, _ in enumerate(lst):
				self.accounts_meta_data[module_name].insert(index, {"auth_fails": 0, "rate_limits": 0, "captcha": 0})

	def get_module_name(self, module_origin: Origin) -> str:
		module_name = 'youtube' if next((s for s in ("yt", "youtube", "youtu_be") if s in module_origin.value), None) else 'instagram'
		return module_name

	def set_module(self, module_origin: Origin) -> None:
		module_name = self.get_module_name(module_origin)
		self.current_module_name = module_name
		if self.current is None:
			self.current = self.accounts[self.current_module_name][self.account_index[self.current_module_name].value]
		self.current_proxy = self.get_random_account_proxy()

	def next(self) -> dict:
		idx = self.account_index[self.current_module_name].value
		idx += 1
		if idx >= len(self.accounts[self.current_module_name]):
			idx = 0
		self.account_index[self.current_module_name].value = idx
		self.current = self.accounts[self.current_module_name][idx]
		logging.info("Selected account index is '%d'", idx)
		return self.current

	def bump_acc_fail(self, key: str, amount: int = 1) -> int:
		try:
			idx = self.account_index[self.current_module_name].value
			meta_list = self.accounts_meta_data[self.current_module_name]
			if idx >= len(meta_list):
				logging.warning("Index '%d' out of range for module '%s' with length '%d'", idx, self.current_module_name, len(meta_list))
				return 0
			self.accounts_meta_data[self.current_module_name][idx][key] += amount
			return meta_list[idx][key]
		except Exception as e:
			logging.warning("Failed to record fail stats")
			logging.exception(e)
		return 0

	def how_much(self, key: str) -> int:
		idx = self.account_index[self.current_module_name].value
		return self.accounts_meta_data[self.current_module_name][idx][key]

	def get_current(self) -> tuple:
		idx = self.account_index[self.current_module_name].value
		return (idx, self.accounts[self.current_module_name][idx])
	
	def get_current_for_module(self, module_origin: Origin) -> tuple:
		module_name = self.get_module_name(module_origin)
		idx = self.account_index[module_name].value
		return (idx, self.accounts[module_name][idx])

	def get_meta_data(self) -> dict:
		idx = self.account_index[self.current_module_name].value - 1
		return self.accounts_meta_data[self.current_module_name][idx]

	def count_service_accounts(self, mod_name: Origin) -> int:
		module_name = 'youtube' if next((s for s in ("yt", "youtube", "youtu_be") if s in mod_name.value), None) else 'instagram'
		if module_name not in self.accounts:
			return 0
		return len(self.accounts[module_name])
	
	def inc_ig_request_count(self, amount: int = 1) -> None:
		self.ig_request_count.value += amount

	def reset_ig_request_count(self) -> None:
		self.ig_request_count.value = 0

	def get_ig_request_count(self) -> int:
		return self.ig_request_count.value


import os
import time

import logging

import json
import requests

class YtAuth(object):
	TV_CLIENT_ID = "861556708454-d6dlm3lh05idd8npek18k6be8ba3oc68.apps.googleusercontent.com"
	TV_CLIENT_SECRET = "SboVhoG9s0rNafixCSGGKXAT"
	YT_SESSION_FILE_TPL = '/var/warp_beacon/yt_session_%d.json'

	process_start_time = 0
	account_index = 0
	yt_session_file = ""

	def __init__(self, account_index: int) -> None:
		self.account_index = account_index
		self.yt_session_file = self.YT_SESSION_FILE_TPL % account_index

	def fetch_token(self) -> dict:
		result = {"user_code": "", "device_code": "", "verification_url": ""}
		http_code = 0
		response_text = ''
		try:
			logging.info("Fetching YT token ...")
			self.process_start_time = 0
			# Subtracting 30 seconds is arbitrary to avoid potential time discrepencies
			self.process_start_time = int(time.time() - 30)
			data = {
				'client_id': self.TV_CLIENT_ID,
				'scope': 'https://www.googleapis.com/auth/youtube'
			}
			response = requests.post(
				url='https://oauth2.googleapis.com/device/code',
				headers={
					"User-Agent": "Mozilla/5.0",
					"accept-language": "en-US,en",
					"Content-Type": "application/json"
				},
				json=data,
				timeout=int(os.environ.get("YT_TIMEOUT", "30"))
			)

			http_code = response.status_code
			response_text = response.text

			if http_code != 200:
				logging.error("Invalid YT HTTP code: '%d'", http_code)
				logging.info("Request dump: '%s'", str(response.__dict__))
			else:
				response_data = response.json()
				result["verification_url"] = response_data['verification_url']
				result["user_code"] = response_data['user_code']
				result["device_code"] = response_data['device_code']
				logging.info("Fetched YT url '%s' and input code '%s'", result["verification_url"], result['user_code'])
		except Exception as e:
			logging.error("Youtube authorization failed!")
			logging.exception(e)

		if http_code != 200:
			raise ValueError(f"Youtube HTTP response code is {http_code}: {response_text}")

		return result

	def confirm_token(self, device_code: str) -> dict:
		response_data = {}
		http_code = 0
		response_text = ''
		try:
			logging.info("Confirming YT auth token ...")
			self.process_start_time = int(time.time()) - self.process_start_time - 20
			data = {
				'client_id': self.TV_CLIENT_ID,
				'client_secret': self.TV_CLIENT_SECRET,
				'device_code': device_code,
				'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
			}
			response = requests.post(
				url='https://oauth2.googleapis.com/token',
				headers={
					"User-Agent": "Mozilla/5.0",
					"accept-language": "en-US,en",
					"Content-Type": "application/json"
				},
				json=data,
				timeout=int(os.environ.get("YT_TIMEOUT", "30"))
			)

			http_code = response.status_code
			response_text = response.text

			if http_code != 200:
				logging.error("Invalid YT HTTP code: '%d'", http_code)
				logging.info("Request dump: '%s'", str(response.__dict__))
			else:
				response_data = response.json()
				response_data["expires"] = self.process_start_time + int(response_data["expires_in"])
		except Exception as e:
			logging.error("Failed to confirm token!")
			logging.exception(e)

		if http_code != 200:
			raise ValueError(f"Youtube HTTP response code is {http_code}: {response_text}")

		return response_data

	def refresh_token(self, refresh_token: str) -> dict:
		response_data = {}
		http_code = 0
		response_text = ''
		try:
			logging.info("Refreshing YT token ...")
			start_time = int(time.time() - 30)
			data = {
				'client_id': self.TV_CLIENT_ID,
				'client_secret': self.TV_CLIENT_SECRET,
				'grant_type': 'refresh_token',
				'refresh_token': refresh_token
			}
			response = requests.post(
				url='https://oauth2.googleapis.com/token',
				headers={
					"User-Agent": "Mozilla/5.0",
					"accept-language": "en-US,en",
					"Content-Type": "application/json"
				},
				json=data,
				timeout=int(os.environ.get("YT_TIMEOUT", "30"))
			)

			http_code = response.status_code
			response_text = response.text

			if http_code != 200:
				logging.error("Invalid YT HTTP code: '%d'", http_code)
				logging.info("Request dump: '%s'", str(response.__dict__))
			else:
				response_data = response.json()
				response_data["expires"] = start_time + int(response_data["expires_in"])
		except Exception as e:
			logging.error("Failed to refresh YT token")
			logging.exception(e)

		if http_code != 200:
			raise ValueError(f"Youtube HTTP response code is {http_code}: {response_text}")
		
		return response_data

	def safe_write_session(self, token_data: dict) -> bool:
		try:
			tmp_filename = f"{self.yt_session_file}~"
			
			if os.path.exists(tmp_filename):
				os.unlink(tmp_filename)
			
			with open(tmp_filename, "w+", encoding="utf-8") as f:
				f.write(json.dumps(token_data))
			
			if os.path.exists(tmp_filename):
				if os.path.exists(self.yt_session_file):
					os.unlink(self.yt_session_file)
				os.rename(src=tmp_filename, dst=self.yt_session_file)
			return True
		except Exception as e:
			logging.error("Failed to write token!")
			logging.exception(e)

		return False
	
	def store_device_code(self, device_code: str) -> bool:
		try:
			device_code_file = f"/tmp/yt_device_code_acc_{self.account_index}"
			logging.info("Storing device code in file '%s'", device_code_file)
			with open(device_code_file, "w+", encoding="utf-8") as f:
				f.write(device_code.strip())
		except Exception as e:
			logging.error("Failed to store device code!")
			logging.exception(e)
			return False

		return True
	
	def load_device_code(self) -> str:
		device_code = ''
		try:
			device_code_file = f"/tmp/yt_device_code_acc_{self.account_index}"
			logging.info("Loading device code from file '%s'", device_code_file)
			with open(device_code_file, 'r', encoding="utf-8") as f:
				device_code = f.read().strip()
			os.unlink(device_code_file)
		except Exception as e:
			logging.error("Failed to load device code for account #%d", self.account_index)
			logging.exception(e)

		return device_code


import os
import pathlib

import socket
import requests.packages.urllib3.util.connection as urllib3_cn
import multiprocessing

from abc import ABC, abstractmethod
from typing import Callable, Union

from PIL import Image
from pillow_heif import register_heif_opener

import logging

class ScraperAbstract(ABC):
	original_gai_family = None
	send_message_to_admin_func: Callable = lambda: None
	request_yt_auth: Callable = lambda: None
	auth_event = None
	account = None
	account_index = 0
	proxy = None

	def __init__(self, account: tuple, proxy: dict=None) -> None:
		self._gai_lock = multiprocessing.Lock()
		self.account_index = account[0]
		self.account = account[1]
		self.proxy = proxy
		#if self.account.get("force_ipv6", False):
		#	self.force_ipv6()

	#def __del__(self) -> None:
	#	if self.account.get("force_ipv6", False):
	#		self.restore_gai()

	@abstractmethod
	def download(self, url: str) -> bool:
		raise NotImplementedError

	@abstractmethod
	def download_hndlr(self, func: Callable, *args: tuple[str], **kwargs: dict[str]) -> Union[str, dict]:
		raise NotImplementedError

	@staticmethod
	def convert_webp_to_png(src_file: str) -> str:
		try:
			if os.path.exists(src_file):
				path_info = pathlib.Path(src_file)
				old_filename = path_info.stem
				new_filename = "%s_converted.%s" % (old_filename, "png")
				new_filepath = "%s/%s" % (os.path.dirname(src_file), new_filename)
				with Image.open(src_file).convert('RGB') as img:
					img.save(new_filepath, 'png')
				os.unlink(src_file)
				return new_filepath
		except Exception as e:
			logging.error("Failed to convert webp file to png!")
			logging.exception(e)

		return ''

	@staticmethod
	def convert_heic_to_png(src_file: str) -> str:
		try:
			if os.path.exists(src_file):
				register_heif_opener()
				path_info = pathlib.Path(src_file)
				old_filename = path_info.stem
				new_filename = "%s_converted.%s" % (old_filename, "png")
				new_filepath = "%s/%s" % (os.path.dirname(src_file), new_filename)
				with Image.open(src_file).convert('RGB') as img:
					img.save(new_filepath, 'png')
				os.unlink(src_file)
				return new_filepath
		except Exception as e:
			logging.error("Failed to convert webp file to png!")
			logging.exception(e)

		return ''

	def force_ipv6(self) -> None:
		def allowed_gai_family():
			"""
			https://github.com/shazow/urllib3/blob/master/urllib3/util/connection.py
			"""
			family = socket.AF_INET
			if urllib3_cn.HAS_IPV6:
				family = socket.AF_INET6 # force ipv6 only if it is available
			return family
		with self._gai_lock:
			if self.original_gai_family is None:
				self.original_gai_family = urllib3_cn.allowed_gai_family
			logging.info("Forcing IPv6 ...")
			urllib3_cn.allowed_gai_family = allowed_gai_family

	def restore_gai(self) -> None:
		return
		with self._gai_lock:
			if self.original_gai_family:
				logging.info("Restoring normal IP stack ...")
				urllib3_cn.allowed_gai_family = self.original_gai_family
				self.original_gai_family = None